from typing import List, Dict, Set
import numpy as np


class EmptyDatasetException(Exception):
    pass


class Dataset:
    def __init__(self,  r: List[List[List or Set[int or str]]]):
        self.__set_nb_rankings(-1)
        self.__set_nb_elements(-1)
        self.__set_is_complete(None)
        self.__set_with_ties(None)
        if len(r) == 0:
            raise EmptyDatasetException
        # updates previous values with right values for n, m and complete
        self.__set_rankings_and_update_properties(r)

    def __get_rankings(self) -> List[List[List or Set[int or str]]]:
        return self.__rankings

    def __get_nb_elements(self) -> int:
        return self.__nb_elements

    def __get_nb_rankings(self) -> int:
        return self.__nb_rankings

    def __get_is_complete(self) -> bool:
        return self.__is_complete

    def __get_with_ties(self) -> bool:
        return self.__with_ties

    def __set_rankings_and_update_properties(self, rankings: List[List[List or Set[int or str]]]):
        self.__rankings = rankings
        self.__set_nb_rankings(len(rankings))
        self.__set_is_complete(self.__check_if_rankings_complete_and_update_n())

    def __set_nb_elements(self, n: int):
        self.__nb_elements = n

    def __set_nb_rankings(self, m: int):
        self.__nb_rankings = m

    def __set_is_complete(self, complete: bool):
        self.__is_complete = complete

    def __set_with_ties(self, with_ties: bool):
        self.__with_ties = with_ties

    n = property(__get_nb_elements, __set_nb_elements)
    m = property(__get_nb_rankings, __set_nb_rankings)
    rankings = property(__get_rankings, __set_rankings_and_update_properties)
    is_complete = property(__get_is_complete, __set_is_complete)
    with_ties = property(__get_with_ties, __set_with_ties)

    def __str__(self)->str:
        print("ok")
        return str(self.rankings)

    def description(self)->str:
        return "Dataset description:\n\telements:"+str(self.n)+"\n\trankings:"+str(self.m)+"\n\tcomplete:"\
               + str(self.is_complete) + "\n\twith ties: "+str(self.with_ties) + "\n\t"+"rankings:\n"\
               + "\n".join("\t\tr"+str(i+1)+" = "+str(self.rankings[i]) for i in range(len(self.rankings)))

    def __check_if_rankings_complete_and_update_n(self):
        self.with_ties = False
        if len(self.rankings) == 0:
            self.n = 0
            self.m = 0
        else:
            elements = {}

            for ranking in self.rankings:
                nb_elements = 0
                for bucket in ranking:
                    nb_elements += len(bucket)
                    if len(bucket) > 1:
                        self.with_ties = True
                    for element in bucket:
                        if element not in elements:
                            elements[element] = 1
                        else:
                            elements[element] += 1
            self.n = len(elements)
            self.m = len(self.rankings)
            for key in elements.keys():
                if elements[key] != self.m:
                    return False
        return True

    def map_elements_id(self) -> Dict[int or str, int]:
        h = {}
        id_elem = 0
        for ranking in self.rankings:
            for bucket in ranking:
                for elem in bucket:
                    if elem not in h:
                        h[elem] = id_elem
                        id_elem += 1
        return h

    def get_positions(self, elements_id: Dict[str or int, int]) -> np.ndarray:
        n = self.n
        m = self.m
        positions = np.zeros((n, m)) - 1
        id_ranking = 0
        for ranking in self.rankings:
            id_bucket = 0
            for bucket in ranking:
                for elem in bucket:
                    positions[elements_id.get(elem)][id_ranking] = id_bucket
                id_bucket += 1
            id_ranking += 1
        return positions

    def pairs_relative_positions(self, positions: np.ndarray) -> np.ndarray:
        n = self.n
        m = self.m
        matrix = np.zeros((n * n, 6))
        for e1 in range(n-1, -1, -1):
            ind1 = n * e1 + e1
            ind2 = ind1
            for e2 in range(e1-1, -1, -1):
                ind1 -= 1
                ind2 -= n
                a = np.count_nonzero(positions[e1] + positions[e2] == -2)
                b = np.count_nonzero(positions[e1] == positions[e2])
                c = np.count_nonzero(positions[e2] == -1)
                d = np.count_nonzero(positions[e1] == -1)
                e = np.count_nonzero(positions[e1] < positions[e2])
                matrix[ind1] = [e-d+a, b-a, m-(a+b+c+d+e), c-a, d-a, a]
                matrix[ind2] = [e - d + a, b - a, m - (a + b + c + d + e), c - a, d - a, a]

        return matrix

    def unified_rankings(self):
        copy_rankings = []
        elements = set()
        for ranking in self.rankings:
            new_ranking = []
            copy_rankings.append(new_ranking)
            for bucket in ranking:
                new_ranking.append(list(bucket))
                for element in bucket:
                    elements.add(element)

        for ranking in copy_rankings:
            elem_ranking = set(elements)
            for bucket in ranking:
                for element in bucket:
                    elem_ranking.remove(element)
            if len(elem_ranking) > 0:
                ranking.append(list(elem_ranking))
        return copy_rankings

    def unified_dataset(self):
        return Dataset(self.unified_rankings())



from typing import List
from numpy import ndarray, asarray
from corankco.algorithms.median_ranking import MedianRanking
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature


class KwikSortAbs(MedianRanking):

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking=False,
            bench_mode=False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """
        sc = asarray(scoring_scheme.penalty_vectors)

        consensus = []
        elements_translated_target = []
        var = self.prepare_internal_vars(elements_translated_target, dataset.rankings)
        self.kwik_sort(consensus, elements_translated_target, var, sc)
        return Consensus(consensus_rankings=[consensus],
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att={ConsensusFeature.AssociatedAlgorithm: self.get_full_name()}
                         )

    def prepare_internal_vars(self, elements_translated_target: List, rankings: List[List[List[int]]]):
        raise NotImplementedError("The method not implemented")
    # protected abstract U prepareInternalVars(
    # List < V > elementsTranslatedTarget, Collection < List < Collection < T >> > rankings);

    def get_pivot(self, elements: List[int], var):
        raise NotImplementedError("The method not implemented")
    # public abstract V getPivot(List < V > elements, U var);

    def where_should_it_be(self, element: int, pivot: int, elements: List[int], var, scoring_scheme: ndarray):
        raise NotImplementedError("The method not implemented")
    # public abstract int whereShouldItBe(V element, V pivot, List < V > elements, U var);

    def kwik_sort(self, consensus: List[List[int]], elements: List[int], var, scoring_scheme: ndarray):
        after = []
        before = []
        same = []
        pivot = -1
        if len(elements) > 0:
            pivot = self.get_pivot(elements, var)
            same.append(pivot)
        for element in elements:
            if element != pivot:
                pos = self.where_should_it_be(element, pivot, elements, var, scoring_scheme)
                if pos < 0:
                    before.append(element)
                elif pos > 0:
                    after.append(element)
                else:
                    same.append(element)

        if len(before) == 1:
            consensus.append(before)
        elif len(before) > 0:
            self.kwik_sort(consensus, before, var, scoring_scheme)
        if len(same) > 0:
            consensus.append(same)
        if len(after) == 1:
            consensus.append(after)
        elif len(after) > 0:
            self.kwik_sort(consensus, after, var, scoring_scheme)

    def get_full_name(self) -> str:
        return "KwikSortAbs"

    def is_scoring_scheme_relevant_when_incomplete_rankings(self, scoring_scheme: ScoringScheme) -> bool:
        return True



from typing import List, Dict, Tuple, Set, Collection
from corankco.algorithms.median_ranking import MedianRanking
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature
from corankco.kemeny_computation import KemenyScoreFactory
from numpy import zeros, array, ndarray, amin, amax, where, asarray, int32, float64
import bioconsertinc


class BioConsert(MedianRanking):
    def __init__(self, starting_algorithms: Collection[MedianRanking] = None):
        if starting_algorithms is None:
            self.__starting_algorithms = []
        else:
            self.__starting_algorithms = starting_algorithms

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking=False,
            bench_mode=False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """

        sc = asarray(scoring_scheme.penalty_vectors)
        rankings = dataset.rankings

        res = []
        elem_id = {}
        id_elements = {}
        id_elem = 0
        nb_rankings = len(rankings)
        for ranking in rankings:
            for bucket in ranking:
                for element in bucket:
                    if element not in elem_id:
                        elem_id[element] = id_elem
                        id_elements[id_elem] = element
                        id_elem += 1
        nb_elements = len(elem_id)

        positions = BioConsert.__get_positions(rankings, elem_id)
        (departure, dst_res) = self.__departure_rankings(dataset, positions, elem_id, scoring_scheme)

        departure_c = array(departure.flatten(), dtype=int32)

        bioconsertinc.bioconsertinc(
                                    array(positions.flatten(), dtype=int32),
                                    departure_c,
                                    array(sc[0], dtype=float64),
                                    array(sc[1], dtype=float64),
                                    int32(nb_elements),
                                    int32(nb_rankings),
                                    int32(len(departure)),
                                    dst_res,
                                    )
        departure = departure_c.reshape(-1, nb_elements)

        ranking_dict = {}

        lowest_distance = amin(dst_res)
        best_rankings = departure[where(dst_res == lowest_distance)[0]].tolist()
        if return_at_most_one_ranking:
            best_rankings = [best_rankings[-1]]
        distinct_rankings = set()
        for ranking_result in best_rankings:
            st_ranking = str(ranking_result)
            if st_ranking not in distinct_rankings:
                distinct_rankings.add(st_ranking)
                ranking_dict.clear()
                el = 0
                for id_bucket in ranking_result:
                    if id_bucket not in ranking_dict:
                        ranking_dict[id_bucket] = [id_elements.get(el)]
                    else:
                        ranking_dict[id_bucket].append(id_elements.get(el))
                    el += 1

                ranking_list = []
                nb_buckets_ranking_i = len(ranking_dict)
                for id_bucket in range(nb_buckets_ranking_i):
                    ranking_list.append(ranking_dict.get(id_bucket))
                res.append(ranking_list)

        return Consensus(consensus_rankings=res,
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att={ConsensusFeature.KemenyScore: lowest_distance,
                              ConsensusFeature.AssociatedAlgorithm: self.get_full_name()
                              }
                         )

    @staticmethod
    def __get_positions(rankings: List[List[List or Set[int or str]]], elements_id: Dict[int or str, int]) -> ndarray:
        m = len(rankings)
        n = len(elements_id)
        positions = zeros((n, m), dtype=int32) - 1
        id_ranking = 0
        for ranking in rankings:
            id_bucket = 0
            for bucket in ranking:
                for element in bucket:
                    positions[elements_id.get(element)][id_ranking] = id_bucket
                id_bucket += 1
            id_ranking += 1
        return positions

    def __departure_rankings(self, dataset: Dataset, positions: ndarray, elements_id: Dict,
                             scoring_scheme: ScoringScheme) -> Tuple[ndarray, ndarray]:

        dst_ini = []
        dataset_unified = dataset.unified_dataset()
        rankings_unified = dataset_unified.rankings

        if len(self.__starting_algorithms) == 0:
            real_pos = array(positions).transpose()
            distinct_rankings = set()
            list_distinct_id_rankings = []

            i = 0
            for ranking in rankings_unified:
                ranking_array = real_pos[i]
                ranking_array[ranking_array == -1] = amax(ranking_array) + 1
                string_ranking = str(ranking_array)
                if string_ranking not in distinct_rankings:
                    distinct_rankings.add(string_ranking)
                    list_distinct_id_rankings.append(i)

                    dst_ini.append(
                        KemenyScoreFactory.get_kemeny_score(scoring_scheme, ranking, dataset))

                i += 1

            dst_ini.append(KemenyScoreFactory.get_kemeny_score(scoring_scheme, [[*elements_id]], dataset))

            departure = zeros((len(list_distinct_id_rankings)+1, len(elements_id)), dtype=int32)
            departure[:-1] = real_pos[asarray(list_distinct_id_rankings)]
        else:
            m = len(self.__starting_algorithms)
            n = len(elements_id)
            departure = zeros((m, n), dtype=int32) - 1
            id_ranking = 0
            for algo in self.__starting_algorithms:
                cons = algo.compute_consensus_rankings(dataset, scoring_scheme, True).consensus_rankings[0]
                dst_ini.append(KemenyScoreFactory.get_kemeny_score(scoring_scheme, cons, dataset))
                id_bucket = 0
                for bucket in cons:
                    for element in bucket:
                        departure[id_ranking][elements_id.get(element)] = id_bucket
                    id_bucket += 1
                id_ranking += 1

        return departure, array(dst_ini, dtype=float64)

    def get_full_name(self) -> str:
        list_alg = list(self.__starting_algorithms)
        res = "BioConsert with "
        if len(list_alg) == 0:
            res += "input rankings as starters"
        else:
            res += "["
            for alg in list_alg[:-1]:
                res += alg.get_full_name() + ", "
            res += list_alg[-1].get_full_name() + "] as starter algorithms"

        return res

    def is_scoring_scheme_relevant_when_incomplete_rankings(self, scoring_scheme: ScoringScheme) -> bool:
        return True



import operator
from numpy import vdot, count_nonzero, shape, array

from corankco.algorithms.median_ranking import MedianRanking
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature


class CopelandMethod(MedianRanking):

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking=False,
            bench_mode=False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """

        sc = scoring_scheme.penalty_vectors

        res = []
        elem_id = {}
        id_elements = {}
        id_elem = 0
        for ranking in dataset.rankings:
            for bucket in ranking:
                for element in bucket:
                    if element not in elem_id:
                        elem_id[element] = id_elem          # dictionnaire pour retrouver l'id a partir d'un element
                        # (id commence a 0)
                        id_elements[id_elem] = element      # dictionnaire pour retrouver l'element a partir de son id
                        id_elem += 1

        # nb_elements = len(elem_id)

        positions = dataset.get_positions(elem_id)
        n = shape(positions)[0]  # nombre d'elements
        m = shape(positions)[1]  # nombre de classements
        cost_before = sc[0]     # definition des differents couts
        cost_tied = sc[1]
        cost_after = array([cost_before[1], cost_before[0], cost_before[2], cost_before[4], cost_before[3],
                            cost_before[5]])
        id_scores = {}                                      # dictionnaire pour retrouver le score d'un element
        # a partir de son id
        for i in range(0, n, 1):                # initialisation du dictionnaire
            id_scores[i] = 0
        for id_el1 in range(0, n, 1):
            mem = positions[id_el1]             # tableau de rangs de el1
            d = count_nonzero(mem == -1)    # nombre de fois ou seulement el1 est absent
            for id_el2 in range(id_el1 + 1, n, 1):
                a = count_nonzero(mem + positions[id_el2] == -2)  # nombre de fois ou el1 et el2 sont absents
                b = count_nonzero(mem == positions[id_el2])     # nombre de fois ou el1 et el2 sont en egalites
                c = count_nonzero(positions[id_el2] == -1)      # nombre de fois ou seulement el2 est absent
                e = count_nonzero(mem < positions[id_el2])      # nombre de fois ou el1 est avant el2
                relative_positions = array([e - d + a, m - e - b - c + a, b - a, c - a, d - a, a])  # vecteur omega
                put_before = vdot(relative_positions, cost_before)  # cout de placer el1 avant el2
                put_after = vdot(relative_positions, cost_after)    # cout de placer el1 apres el2
                put_tied = vdot(relative_positions, cost_tied)      # cout de placer el1 a egalite avec el2
                if put_before < put_after and put_before < put_tied:
                    id_scores[id_el1] += 1
                    id_scores[id_el2] -= 1
                elif put_after < put_before and put_after < put_tied:
                    id_scores[id_el1] -= 1
                    id_scores[id_el2] += 1
        sorted_ids = CopelandMethod.sorted_dictionary_keys(id_scores)  # liste des cles du dictionnaire trie par
        # valeurs decroissantes
        bucket = []
        previous_id = sorted_ids[0]
        for id_elem in sorted_ids:
            if id_scores.get(previous_id) == id_scores.get(id_elem):  # si l'elem actuel a le meme score que l'element
                # precedent
                bucket.append(id_elements.get(id_elem))                  # on le place dans le meme bucket que celui ci
            else:
                res.append(bucket)                                  # sinon, on concatene le bucket a la liste resultat
                bucket = [id_elements.get(id_elem)]                 # on reinitialise le bucket avec le candidat actuel
            previous_id = id_elem
        res.append(bucket)                  # on concatene le dernier bucket a la liste resultat
        return Consensus(consensus_rankings=[res],
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att={
                              ConsensusFeature.AssociatedAlgorithm: self.get_full_name()
                              }
                         )

    # tri du dictionnaire dans l'ordre decroissant des scores, tri "timsort" par defaut. Complexite O(nlog n)
    # au pire des cas
    # retourne la liste des cles ordonnées par valeurs decroissantes
    @staticmethod
    def sorted_dictionary_keys(d):
        d = (sorted(d.items(), key=operator.itemgetter(1), reverse=True))
        res = []
        for (k, v) in d:
            res.append(k)
        return res

    def get_full_name(self) -> str:
        return "CopelandMethod"

    def is_scoring_scheme_relevant_when_incomplete_rankings(self, scoring_scheme: ScoringScheme) -> bool:
        return True



from enum import Enum, unique
from corankco.algorithms.allTied.alltied import AllTied
from corankco.algorithms.bioconsert.bioconsert import BioConsert
from corankco.algorithms.parcons.parcons import ParCons
from corankco.algorithms.exact.exactalgorithm import ExactAlgorithm
from corankco.algorithms.kwiksort.kwiksortrandom import KwikSortRandom
from corankco.algorithms.repeatchoice.repeatchoice import RepeatChoice
from corankco.algorithms.pickaperm.pickaperm import PickAPerm
from corankco.algorithms.medrank.medrank import MedRank
from corankco.algorithms.borda.borda import BordaCount
from corankco.algorithms.bioconsert.bioco import BioCo
from corankco.algorithms.copeland.copeland import CopelandMethod


class AlgorithmEnumeration:
    median_ranking_algorithms = [
        AllTied,
        BioConsert,
        ParCons,
        ExactAlgorithm,
        KwikSortRandom,
        RepeatChoice,
        PickAPerm,
        MedRank,
        BordaCount,
        BioCo,
        CopelandMethod
    ]


@unique
class Algorithm(Enum):
    AllTied = 0
    BioConsert = 1
    ParCons = 2
    Exact = 3
    KwikSortRandom = 4
    RepeatChoice = 5
    PickAPerm = 6
    MedRank = 7
    BordaCount = 8
    BioCo = 9
    CopelandMethod = 10

    @staticmethod
    def get_all():
        return [Algorithm.AllTied, Algorithm.BioConsert, Algorithm.ParCons,
                Algorithm.Exact, Algorithm.KwikSortRandom, Algorithm.RepeatChoice, Algorithm.PickAPerm,
                Algorithm.MedRank, Algorithm.BordaCount, Algorithm.BioCo, Algorithm.CopelandMethod]
    @staticmethod
    def get_all_compatible_with_any_scoring_scheme():
        return [Algorithm.BioConsert, Algorithm.ParCons, Algorithm.Exact,
                Algorithm.KwikSortRandom, Algorithm.CopelandMethod]





from typing import List, Dict, Tuple, Set
from corankco.algorithms.median_ranking import MedianRanking
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature
from numpy import ndarray, array, shape, zeros, count_nonzero, vdot, asarray
from operator import itemgetter
import cplex


class ExactAlgorithmCplex(MedianRanking):
    def __init__(self, limit_time_sec=0, scoring_scheme=None):
        if limit_time_sec > 0:
            self.__limit_time_sec = limit_time_sec
        else:
            self.__limit_time_sec = 0
        self.__scoring_scheme = scoring_scheme

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking=False,
            bench_mode=False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """

        rankings = dataset.rankings
        elem_id = {}
        id_elements = {}
        id_elem = 0
        for ranking in rankings:
            for bucket in ranking:
                for element in bucket:
                    if element not in elem_id:
                        elem_id[element] = id_elem
                        id_elements[id_elem] = element
                        id_elem += 1
        nb_elem = len(elem_id)

        positions = ExactAlgorithmCplex.__positions(rankings, elem_id)

        sc = asarray(scoring_scheme.penalty_vectors)

        mat_score, ties_must_be_checked = self.__cost_matrix(positions, asarray(sc))
        map_elements_cplex = {}
        my_prob = cplex.Cplex()  # initiate
        my_prob.set_results_stream(None)  # mute
        my_prob.parameters.mip.tolerances.mipgap.set(0.0)
        my_prob.parameters.mip.pool.absgap.set(0.0)
        if self.__limit_time_sec > 0:
            my_prob.parameters.tuning.timelimit.set(self.__limit_time_sec)
        my_prob.objective.set_sense(my_prob.objective.sense.minimize)  # we want to minimize the objective function
        if not return_at_most_one_ranking:
            my_prob.parameters.mip.pool.intensity.set(4)
            my_prob.parameters.mip.limits.populate.set(10000000)

        my_obj = []
        my_ub = []
        my_lb = []
        my_names = []

        cpt = 0
        for i in range(nb_elem):
            for j in range(nb_elem):
                if not i == j:
                    s = "x_%s_%s" % (i, j)
                    my_obj.append(mat_score[i][j][0])
                    my_ub.append(1.0)
                    my_lb.append(0.0)
                    my_names.append(s)
                    map_elements_cplex[cpt] = ("x", i, j)
                    cpt += 1

        for i in range(nb_elem):
            for j in range(i+1, nb_elem):
                s = "t_%s_%s" % (i, j)
                my_obj.append(mat_score[i][j][2])
                my_ub.append(1.0)
                my_lb.append(0.0)
                my_names.append(s)
                map_elements_cplex[cpt] = ("t", i, j)
                cpt += 1
        my_prob.variables.add(obj=my_obj, lb=my_lb, ub=my_ub, types="B"*cpt, names=my_names)

        # rhs = right hand side
        my_rhs = []
        my_rownames = []

        # inequations : E for Equality, G for >=  and L for <=
        my_sense = "E" * int(nb_elem*(nb_elem-1)/2) + "L" * (3*nb_elem * (nb_elem-1) * (nb_elem-2))

        rows = []

        # add the binary order constraints
        count = 0
        for i in range(0, nb_elem - 1):
            for j in range(i + 1, nb_elem):
                if not i == j:
                    s = "c%s" % count
                    count += 1
                    my_rhs.append(1)
                    my_rownames.append(s)
                    first_var = "x_%s_%s" % (i, j)
                    second_var = "x_%s_%s" % (j, i)
                    third_var = "t_%s_%s" % (i, j)

                    row = [[first_var, second_var, third_var], [1.0, 1.0, 1.0]]
                    rows.append(row)
        # add the transitivity constraints
        for i in range(0, nb_elem):
            for j in range(nb_elem):
                if j != i:
                    i_bef_j = "x_%s_%s" % (i, j)
                    if i < j:
                        i_tie_j = "t_%s_%s" % (i, j)
                    else:
                        i_tie_j = "t_%s_%s" % (j, i)
                    for k in range(nb_elem):
                        if k != i and k != j:
                            my_rownames.append("c%s" % count)
                            my_rhs.append(1)
                            count += 1
                            if j < k:
                                j_tie_k = "t_%s_%s" % (j, k)
                            else:
                                j_tie_k = "t_%s_%s" % (k, j)
                            rows.append([[i_bef_j, "x_%s_%s" % (j, k), j_tie_k, "x_%s_%s" % (i, k)], [1., 1., 1., -1.]])

                            my_rownames.append("c%s" % count)
                            my_rhs.append(1)
                            count += 1
                            rows.append([[i_bef_j, i_tie_j, "x_%s_%s" % (j, k), "x_%s_%s" % (i, k)], [1., 1., 1., -1.]])

                            if i < k:
                                i_tie_k = "t_%s_%s" % (i, k)
                            else:
                                i_tie_k = "t_%s_%s" % (k, i)

                            my_rownames.append("c%s" % count)
                            my_rhs.append(3)
                            count += 1
                            rows.append([[i_tie_j, j_tie_k, i_tie_k], [2.0, 2.0, -1.0]])

        # if tie is not the single best choice for any pair x,y, then there is a permutation median
        if not ties_must_be_checked:
            my_sense += "E" * int(nb_elem*(nb_elem-1)/2)
            for i in range(0, nb_elem - 1):
                for j in range(i + 1, nb_elem):
                    if not i == j:
                        s = "c%s" % count
                        count += 1
                        my_rhs.append(0)
                        my_rownames.append(s)
                        row = [["t_%s_%s" % (i, j)], [1]]
                        rows.append(row)

        my_prob.linear_constraints.add(lin_expr=rows, senses=my_sense, rhs=my_rhs, names=my_rownames)
        medianes = []

        if not return_at_most_one_ranking:
            my_prob.populate_solution_pool()

            nb_optimal_solutions = my_prob.solution.pool.get_num()
            for i in range(nb_optimal_solutions):
                names = my_prob.solution.pool.get_values(i)
                medianes.append(ExactAlgorithmCplex.__create_consensus(nb_elem, names, map_elements_cplex, id_elements))
        else:
            my_prob.solve()
            x = my_prob.solution.get_values()
            medianes.append(ExactAlgorithmCplex.__create_consensus(nb_elem, x, map_elements_cplex, id_elements))

        return Consensus(consensus_rankings=medianes,
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att={ConsensusFeature.IsNecessarilyOptimal: True,
                              ConsensusFeature.KemenyScore: my_prob.solution.get_objective_value(),
                              ConsensusFeature.AssociatedAlgorithm: self.get_full_name()
                              })

    @staticmethod
    def __create_consensus(nb_elem: int, x: List, map_elements_cplex: Dict, id_elements: Dict):
        ranking = []
        count_after = {}
        for i in range(nb_elem):
            count_after[i] = 0
        for var in range(len(x)):
            if abs(x[var] - 1) < 0.001:
                tple = map_elements_cplex[var]
                if tple[0] == "x":
                    count_after[tple[2]] += 1

        current_nb_def = 0
        bucket = []
        for elem, nb_defeats in (sorted(count_after.items(), key=itemgetter(1))):
            if nb_defeats == current_nb_def:
                bucket.append(id_elements.get(elem))
            else:
                ranking.append(bucket)
                bucket = [id_elements.get(elem)]
                current_nb_def = nb_defeats
        ranking.append(bucket)
        return ranking

    @staticmethod
    def __cost_matrix(positions: ndarray, matrix_scoring_scheme: ndarray) -> Tuple[ndarray, bool]:
        cost_before = matrix_scoring_scheme[0]
        cost_tied = matrix_scoring_scheme[1]
        cost_after = array([cost_before[1], cost_before[0], cost_before[2], cost_before[4], cost_before[3],
                            cost_before[5]])
        n = shape(positions)[0]
        m = shape(positions)[1]

        matrix = zeros((n, n, 3))
        ties_must_be_checked = False
        for e1 in range(n):
            mem = positions[e1]
            d = count_nonzero(mem == -1)
            for e2 in range(e1 + 1, n):
                a = count_nonzero(mem + positions[e2] == -2)
                b = count_nonzero(mem == positions[e2])
                c = count_nonzero(positions[e2] == -1)
                e = count_nonzero(mem < positions[e2])
                relative_positions = array([e - d + a, m - e - b - c + a, b - a, c - a, d - a, a])
                put_before = vdot(relative_positions, cost_before)
                put_after = vdot(relative_positions, cost_after)
                put_tied = vdot(relative_positions, cost_tied)
                matrix[e1][e2] = [put_before, put_after, put_tied]
                matrix[e2][e1] = [put_after, put_before, put_tied]
                if put_tied <= put_after and put_tied <= put_before:
                    ties_must_be_checked = True

        return matrix, ties_must_be_checked

    @staticmethod
    def __positions(rankings: List[List[List or Set[int or str]]], elements_id: Dict) -> ndarray:
        positions = zeros((len(elements_id), len(rankings)), dtype=int) - 1
        id_ranking = 0
        for ranking in rankings:
            id_bucket = 0
            for bucket in ranking:
                for element in bucket:
                    positions[elements_id.get(element)][id_ranking] = id_bucket
                id_bucket += 1
            id_ranking += 1
        return positions

    def get_full_name(self) -> str:
        return "Exact algorithm ILP Cplex"

    def is_scoring_scheme_relevant_when_incomplete_rankings(self, scoring_scheme: ScoringScheme) -> bool:
        return True



from corankco.algorithms.median_ranking import MedianRanking, ScoringSchemeNotHandledException
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature


class BordaCount(MedianRanking):
    def __init__(self,  use_bucket_id=False):
        self.useBucketIdAndNotBucketSize = use_bucket_id

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking=False,
            bench_mode=False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """

        if not dataset.is_complete and not self.is_scoring_scheme_relevant_when_incomplete_rankings(scoring_scheme):
            raise ScoringSchemeNotHandledException

        if scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme().penalty_vectors) or \
                scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme_p(0.5).penalty_vectors):
            rankings_to_use = dataset.unified_rankings()
        else:
            rankings_to_use = dataset.rankings

        points = {}
        for ranking in rankings_to_use:
            id_bucket = 1
            for bucket in ranking:
                for elem in bucket:
                    if elem not in points:
                        points[elem] = {}
                        points[elem][0] = 0
                        points[elem][1] = 0

                    points[elem][0] += id_bucket
                    points[elem][1] += 1
                if self.useBucketIdAndNotBucketSize:
                    id_bucket += 1
                else:
                    id_bucket += len(bucket)
        lis = []
        for elem in points.keys():
            lis.append((elem, points[elem][0] * 1.0 / points[elem][1]))
        tri = sorted(lis, key=lambda col: col[1])
        consensus = []
        bucket = []
        last = -1
        for duo in tri:
            if duo[1] != last:
                last = duo[1]
                bucket = []
                consensus.append(bucket)
            bucket.append(duo[0])
        return Consensus(consensus_rankings=[consensus],
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att={
                              ConsensusFeature.AssociatedAlgorithm: self.get_full_name()
                              }
                         )

    def get_full_name(self) -> str:
        return "BordaCount"

    def is_scoring_scheme_relevant_when_incomplete_rankings(self, scoring_scheme: ScoringScheme) -> bool:
        return scoring_scheme.is_equivalent_to(ScoringScheme.get_induced_measure_scoring_scheme().penalty_vectors) or \
               scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme().penalty_vectors) or \
               scoring_scheme.is_equivalent_to(ScoringScheme.get_induced_measure_scoring_scheme_p(0.5).penalty_vectors)\
               or scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme_p(0.5).penalty_vectors)



from itertools import zip_longest

from corankco.algorithms.median_ranking import MedianRanking, ScoringSchemeNotHandledException
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature


class MedRank(MedianRanking):
    def __init__(self,  h=0.5):
        if h < 0:
            h = 0
        elif h > 1:
            h = 1
        self.h = h

    # Complexity : 0 (2 * n) with adaptation for induced measure
    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking=False,
            bench_mode=False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """

        if not dataset.is_complete and not self.is_scoring_scheme_relevant_when_incomplete_rankings(scoring_scheme):
            raise ScoringSchemeNotHandledException

        if scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme().penalty_vectors) or \
                scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme_p(0.5).penalty_vectors):
            rankings_to_use = dataset.unified_rankings()
        else:
            rankings_to_use = dataset.rankings
        has = {}

        nb_rankings_needed = {}
        already_put = set()

        for ranking in rankings_to_use:
            for bucket in ranking:
                for element in bucket:
                    if element not in nb_rankings_needed:
                        nb_rankings_needed[element] = self.h
                    else:
                        nb_rankings_needed[element] += self.h

        bucket_res = []
        ranking_res = []

        for reorganized in zip_longest(*rankings_to_use):
            for bucket in reorganized:
                if bucket is not None:
                    for element in bucket:
                        if element not in already_put:
                            if element not in has:
                                has[element] = 1
                                if nb_rankings_needed[element] <= 1:
                                    bucket_res.append(element)
                                    already_put.add(element)
                            else:
                                has[element] += 1
                                if has[element] >= nb_rankings_needed[element]:
                                    bucket_res.append(element)
                                    already_put.add(element)
            if len(bucket_res) > 0:
                ranking_res.append(bucket_res)
                bucket_res = []

        rankings_consensus = [ranking_res] if len(ranking_res) > 0 else [[]]
        return Consensus(consensus_rankings=rankings_consensus,
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att={
                              ConsensusFeature.AssociatedAlgorithm: self.get_full_name()
                              }
                         )

    def get_full_name(self) -> str:
        return "MedRank"

    def is_scoring_scheme_relevant_when_incomplete_rankings(self, scoring_scheme: ScoringScheme) -> bool:
        return scoring_scheme.is_equivalent_to(ScoringScheme.get_induced_measure_scoring_scheme().penalty_vectors) or \
               scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme().penalty_vectors) or \
               scoring_scheme.is_equivalent_to(ScoringScheme.get_induced_measure_scoring_scheme_p(0.5).penalty_vectors)\
               or scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme_p(0.5).penalty_vectors)



from typing import List
from numpy import zeros, count_nonzero, vdot, ndarray
from random import choice
from corankco.algorithms.kwiksort.kwiksortabs import KwikSortAbs


class KwikSortRandom(KwikSortAbs):

    def prepare_internal_vars(self, elements_translated_target: List, rankings: List[List[List[int]]]):
        elements = {}
        id_element = 0
        for ranking in rankings:
            for bucket in ranking:
                for element in bucket:
                    if element not in elements:
                        elements_translated_target.append(element)
                        elements[element] = id_element
                        id_element += 1

        positions = zeros((len(elements), len(rankings)), dtype=int)-1

        id_ranking = 0
        for ranking in rankings:
            id_bucket = 0
            for bucket in ranking:
                for element in bucket:
                    positions[elements.get(element)][id_ranking] = id_bucket
                id_bucket += 1
            id_ranking += 1
        tup = (elements, positions)
        return tup

    def get_pivot(self, elements: List[int], var):
        return choice(elements)

    def where_should_it_be(self, element: int, pivot: int, elements: List[int], var, scoring_scheme: ndarray):
        pivot_var = var[1][var[0].get(pivot)]
        element_var = var[1][var[0].get(element)]
        m = len(pivot_var)

        a = count_nonzero(pivot_var + element_var == -2)
        b = count_nonzero(pivot_var == element_var)
        c = count_nonzero(pivot_var == -1)
        d = count_nonzero(element_var == -1)
        e = count_nonzero(element_var < pivot_var)

        comp = [e-d+a, m-e-b-c+a, b-a, c-a, d-a, a]
        cost_before = vdot(scoring_scheme[0], comp)
        cost_same = vdot(scoring_scheme[1], comp)
        cost_after = vdot(scoring_scheme[0], [m-e-b-c+a,  e-d+a, b-a, d-a, c-a, a])

        if cost_same <= cost_before:
            if cost_same <= cost_after:
                return 0
            return 1
        else:
            if cost_before <= cost_after:
                return -1
        return 1

    def get_full_name(self) -> str:
        return "KwikSortRandom"



from typing import List
from math import isnan


class InvalidScoringScheme(Exception):
    pass


class NotRelevantScoringScheme(Exception):
    pass


class ScoringScheme:

    def __init__(self, penalties=None):
        if penalties is None:
            self._penalties = [[0., 1., 1., 0., 1., 0.], [1., 1., 0., 1., 1., 0.]]
        else:
            if type(penalties) is not list or len(penalties) != 2 or len(penalties[0]) != 6 or len(penalties[1]) != 6:
                raise InvalidScoringScheme
            penalties_copy = [[], []]
            for pen in penalties[0]:
                penalties_copy[0].append(float(pen))
            for pen in penalties[1]:
                penalties_copy[1].append(float(pen))

            if penalties_copy[1][0] != penalties_copy[1][1] or penalties_copy[1][3] != penalties_copy[1][4]:
                raise NotRelevantScoringScheme
            self._penalties = penalties_copy

    @property
    def penalty_vectors(self) -> List[List[float]]:
        return self._penalties

    def __str__(self) -> str:
        return str(self._penalties)

    def description(self):
        return "\nScoring scheme description\n\tx before y in consensus\n\t\tx before y in input ranking: "\
               + str(self._penalties[0][0])\
               + "\n\t\ty before x in input ranking: "+str(self._penalties[0][1])\
               + "\n\t\tx and y tied in input ranking: " + str(self._penalties[0][2])\
               + "\n\t\tx present y missing in input ranking: " + str(self._penalties[0][3])\
               + "\n\t\tx missing y present ranking: " + str(self._penalties[0][4]) \
               + "\n\t\tx and y missing in input ranking: " + str(self._penalties[0][5]) \
               + "\n\tx and y tied in consensus\n\t\tx before y in input ranking: " + str(self._penalties[1][0]) \
                + "\n\t\ty before x in input ranking: " + str(self._penalties[1][1]) \
                + "\n\t\tx and y tied in input ranking: " + str(self._penalties[1][2]) \
                + "\n\t\tx present y missing in input ranking: " + str(self._penalties[1][3]) \
                + "\n\t\tx missing y present ranking: " + str(self._penalties[1][4]) \
                + "\n\t\tx and y missing in input ranking: " + str(self._penalties[1][5])

    @staticmethod
    def get_pseudodistance_scoring_scheme():
        return ScoringScheme.get_pseudodistance_scoring_scheme_p(1.)

    @staticmethod
    def get_unifying_scoring_scheme():
        return ScoringScheme.get_unifying_scoring_scheme_p(1.)

    @staticmethod
    def get_induced_measure_scoring_scheme():
        return ScoringScheme.get_induced_measure_scoring_scheme_p(1.)

    @staticmethod
    def get_pseudodistance_scoring_scheme_p(p: float):
        return ScoringScheme(penalties=[[0., 1., p, 0., 1., 0.], [p, p, 0., p, p, 0.]])

    @staticmethod
    def get_unifying_scoring_scheme_p(p: float):
        return ScoringScheme(penalties=[[0., 1., p, 0., 1., p], [p, p, 0., p, p, 0.]])

    @staticmethod
    def get_induced_measure_scoring_scheme_p(p: float):
        return ScoringScheme(penalties=[[0., 1., p, 0., 0., 0.], [p, p, 0., 0., 0., 0.]])

    @staticmethod
    def get_fagin_score_for_complete_ranking(p: float):
        return ScoringScheme(penalties=[[0., 1., p, 0., 0., 0.], [p, p, 0., 0., 0., 0.]])

    def is_equivalent_to(self, sc2: List[List[float or int]]) -> bool:
        return self.__is_equivalent_to_generic(sc2, 6)

    def is_equivalent_to_on_complete_rankings_only(self, sc2: List[List[float or int]]) -> bool:
        return self.__is_equivalent_to_generic(sc2, 3)

    def __is_equivalent_to_generic(self, sc2: List[List[float or int]], stop: int) -> bool:
        second_scoring_scheme = ScoringScheme(sc2)
        pen2 = second_scoring_scheme._penalties
        pen1 = self._penalties
        coefficient = float("nan")
        for i in range(stop):
            if pen1[0][i] == 0:
                if pen2[0][i] != 0:
                    return False
            else:
                if pen2[0][i] == 0:
                    return False
                else:
                    if isnan(coefficient):
                        coefficient = pen1[0][i] / pen2[0][i]
                    else:
                        if pen1[0][i] / pen2[0][i] != coefficient:
                            return False
        return True



from corankco.consensus import Consensus
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.algorithms.enumeration import AlgorithmEnumeration, Algorithm


class KemRankAgg:
    def __init__(self):
        pass

    @staticmethod
    def compute_consensus(dataset: Dataset,
                          scoring_scheme: ScoringScheme,
                          algorithm: Algorithm,
                          single_consensus=False) -> Consensus:
        alg = AlgorithmEnumeration.median_ranking_algorithms[algorithm.value]
        return alg().compute_consensus_rankings(dataset, scoring_scheme, single_consensus)



from corankco.algorithms.bioconsert.bioconsert import BioConsert
from corankco.algorithms.borda.borda import BordaCount


class BioCo(BioConsert):
    def __init__(self):
        super().__init__(starting_algorithms=[BordaCount()])

    def get_full_name(self) -> str:
        return "Bioco ( " + super().get_full_name() + ")"



from random import shuffle
from numpy import zeros, argmax
from functools import cmp_to_key

from corankco.algorithms.median_ranking import MedianRanking, ScoringSchemeNotHandledException
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature


class RepeatChoice(MedianRanking):

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking=False,
            bench_mode=False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """

        if not dataset.is_complete and not self.is_scoring_scheme_relevant_when_incomplete_rankings(scoring_scheme):
            raise ScoringSchemeNotHandledException

        if scoring_scheme.is_equivalent_to([[0, 1, 1, 0, 1, 1], [1, 1, 0, 1, 1, 0]]):
            rankings_to_use = dataset.unified_rankings()
        else:
            rankings_to_use = dataset.rankings

        nb_rankings = len(rankings_to_use)
        rankings_copy = list(rankings_to_use)
        shuffle(rankings_copy)
        h = {}
        id_ranking = 0
        for ranking in rankings_copy:
            id_bucket = 0
            for bucket in ranking:
                for element in bucket:
                    if element not in h:
                        h[element] = zeros(nb_rankings, dtype=int) - 1
                    h[element][id_ranking] = id_bucket
                id_bucket += 1
            id_ranking += 1

        res = []
        for el in sorted(h.items(), key=cmp_to_key(RepeatChoice.__compare)):
            res.append([el[0]])

        # kem = KemenyComputingFactory(scoring_scheme=self.scoring_scheme)
        # kem = KendallTauGeneralizedNlogN()
        return Consensus(consensus_rankings=[res],
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att={
                             ConsensusFeature.AssociatedAlgorithm: self.get_full_name()
                         }
                         )

    @staticmethod
    def __compare(e1: tuple, e2: tuple) -> int:
        first_ind_array_e1_inf_array_e2 = argmax(e1[1] < e2[1])
        first_ind_array_e2_inf_array_e1 = argmax(e2[1] < e1[1])
        if first_ind_array_e1_inf_array_e2 < first_ind_array_e2_inf_array_e1:
            return -1
        elif first_ind_array_e2_inf_array_e1 < first_ind_array_e1_inf_array_e2:
            return 1
        return 0

    def get_full_name(self) -> str:
        return "RepeatChoice"

    def is_scoring_scheme_relevant_when_incomplete_rankings(self, scoring_scheme: ScoringScheme) -> bool:
        return scoring_scheme.is_equivalent_to(ScoringScheme.get_induced_measure_scoring_scheme().penalty_vectors) or \
               scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme().penalty_vectors) or \
               scoring_scheme.is_equivalent_to(ScoringScheme.get_induced_measure_scoring_scheme_p(0.5).penalty_vectors)\
               or scoring_scheme.is_equivalent_to(ScoringScheme.get_unifying_scoring_scheme_p(0.5).penalty_vectors)



from abc import ABC
from typing import List

from corankco.algorithms.median_ranking import MedianRanking
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature
from corankco.kemeny_computation import KemenyScoreFactory


class InompleteRankingsIncompatibleWithScoringSchemeException(Exception):
    pass


class PickAPerm(MedianRanking):

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking=False,
            bench_mode=False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """
        sc = scoring_scheme.penalty_vectors
        if not dataset.is_complete:
            for i in range(3):
                if sc[0][i] > sc[0][i+3] or sc[1][i] > sc[1][i+3]:
                    raise InompleteRankingsIncompatibleWithScoringSchemeException
            rankings_to_use = dataset.unified_rankings()
        else:
            rankings_to_use = dataset.rankings

        k = KemenyScoreFactory()
        dst_min = float('inf')
        consensus = [[]]
        for ranking in rankings_to_use:
            dist = k.get_kemeny_score(scoring_scheme, ranking, dataset)
            if dist < dst_min:
                dst_min = dist
                consensus.clear()
                consensus.append(ranking)
            elif dist == dst_min and not return_at_most_one_ranking:
                consensus.append(ranking)

        return Consensus(consensus_rankings=consensus,
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att={ConsensusFeature.KemenyScore: dst_min,
                              ConsensusFeature.AssociatedAlgorithm: self.get_full_name()
                              }
                         )

    def get_full_name(self) -> str:
        return "Pick a Perm"

    def is_scoring_scheme_relevant_when_incomplete_rankings(self, scoring_scheme: ScoringScheme) -> bool:
        return True



from corankco.algorithms.median_ranking import MedianRanking
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature
from corankco.algorithms.bioconsert.bioconsert import BioConsert
from corankco.algorithms.exact.exactalgorithm import ExactAlgorithm
from typing import Tuple
from itertools import combinations
from numpy import vdot, ndarray, count_nonzero, shape, array, zeros, asarray
from igraph import Graph


class ParCons(MedianRanking):
    def __init__(self, algorithm_to_complete=None, bound_for_exact=80):
        if isinstance(algorithm_to_complete, MedianRanking):
            self.alg = algorithm_to_complete
        else:
            self.alg = BioConsert(starting_algorithms=None)
        self.bound_for_exact = bound_for_exact

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking=False,
            bench_mode=False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """
        optimal = True
        sc = asarray(scoring_scheme.penalty_vectors)
        rankings = dataset.rankings
        res = []
        elem_id = {}
        id_elements = {}
        id_elem = 0
        for ranking in rankings:
            for bucket in ranking:
                for element in bucket:
                    if element not in elem_id:
                        elem_id[element] = id_elem
                        id_elements[id_elem] = element
                        id_elem += 1

        positions = dataset.get_positions(elem_id)
        gr1, mat_score = self.__graph_of_elements(positions, sc)
        scc = gr1.components()
        for scc_i in scc:
            if len(scc_i) == 1:
                res.append([id_elements.get(scc_i[0])])
            else:
                all_tied = True
                for e1, e2 in combinations(scc_i, 2):
                    if mat_score[e1][e2][2] > mat_score[e1][e2][0] or mat_score[e1][e2][2] > mat_score[e1][e2][1]:
                        all_tied = False
                        break
                if all_tied:
                    buck = []
                    for el in scc_i:
                        buck.append(id_elements.get(el))
                    res.append(buck)
                else:
                    set_scc = set(scc_i)
                    project_rankings = []
                    for ranking in rankings:
                        project_ranking = []
                        for bucket in ranking:
                            project_bucket = []
                            for elem in bucket:
                                if elem_id.get(elem) in set_scc:
                                    project_bucket.append(elem)
                            if len(project_bucket) > 0:
                                project_ranking.append(project_bucket)
                        if len(project_ranking) > 0:
                            project_rankings.append(project_ranking)
                    if len(scc_i) > self.bound_for_exact:
                        cons_ext = self.alg.compute_consensus_rankings(Dataset(project_rankings),
                                                                       scoring_scheme,
                                                                       True).consensus_rankings[0]
                        res.extend(cons_ext)
                        optimal = False
                    else:
                        cons_ext = ExactAlgorithm().compute_consensus_rankings(Dataset(project_rankings),
                                                                               scoring_scheme,
                                                                               True).consensus_rankings[0]
                        res.extend(cons_ext)
        hash_information = {ConsensusFeature.IsNecessarilyOptimal: optimal,
                            ConsensusFeature.AssociatedAlgorithm: self.get_full_name()
                            }
        if not bench_mode:
            cfc_name = []
            for scc_i in scc:
                group = set()
                for elem in scc_i:
                    group.add(id_elements[elem])
                cfc_name.append(group)
            hash_information[ConsensusFeature.WeakPartitioning] = str(cfc_name)

        return Consensus(consensus_rankings=[res],
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att=hash_information)

    @staticmethod
    def __graph_of_elements(positions: ndarray, matrix_scoring_scheme: ndarray) -> Tuple[Graph, ndarray]:
        graph_of_elements = Graph(directed=True)
        cost_before = matrix_scoring_scheme[0]
        cost_tied = matrix_scoring_scheme[1]
        cost_after = array([cost_before[1], cost_before[0], cost_before[2], cost_before[4], cost_before[3],
                            cost_before[5]])
        n = shape(positions)[0]
        m = shape(positions)[1]
        for i in range(n):
            graph_of_elements.add_vertex(name=str(i))

        matrix = zeros((n, n, 3))
        edges = []
        for e1 in range(n):
            mem = positions[e1]
            d = count_nonzero(mem == -1)
            for e2 in range(e1 + 1, n):
                a = count_nonzero(mem + positions[e2] == -2)
                b = count_nonzero(mem == positions[e2])
                c = count_nonzero(positions[e2] == -1)
                e = count_nonzero(mem < positions[e2])
                relative_positions = array([e - d + a, m - e - b - c + a, b - a, c - a, d - a, a])
                put_before = vdot(relative_positions, cost_before)
                put_after = vdot(relative_positions, cost_after)
                put_tied = vdot(relative_positions, cost_tied)
                if put_before > put_after or put_before > put_tied:
                    edges.append((e2, e1))
                if put_after > put_before or put_after > put_tied:
                    edges.append((e1, e2))
                matrix[e1][e2] = [put_before, put_after, put_tied]
                matrix[e2][e1] = [put_after, put_before, put_tied]
        graph_of_elements.add_edges(edges)
        return graph_of_elements, matrix

    def get_full_name(self) -> str:
        return "ParCons, uses  \"" + self.alg.get_full_name() + "\" on groups of size >  " + str(self.bound_for_exact)

    def is_scoring_scheme_relevant_when_incomplete_rankings(self, scoring_scheme: ScoringScheme) -> bool:
        return True
